/*=============================================================================== 
                    MULTI-FUNCTION ANIMATION DECODER: SCHEDULER VERSION 4
  MFAS: Multi-Function Animation Scheduler (MiniPro) -  F.Miller 4/26/16
  ===============================================================================
  This sketch runs a 24-hour "schedule" of LocoNet functions (0-8) for
  any LocoNet address.  Functions are turned ON and OFF by schedule entry.

  The clock time is displayed in Hours, Minutes and AM/PM but the in-memory
  schedule uses 24 hour times of hours and minutes. (Hours 1-24, minutes 0-59)
  
  The command buttons on LCD are used as follows:
  left=Start; right=Stop/Restart; top=faster clock (up to 20x actual); bottom=slower
  clock (down to 6x actual); upper-right=hold (while pushed); far right=reset system.
  Note it may be necessary to push these buttons several times to catch cycle.
  
  At startup, the unit will display time-speed setting (120x actual) then wait at
  time 01:00 AM for "start" button.
  
  This version uses a MiniPro and a 4-bit parallel LCD Shield for time display.
  The LCD Shield pins have been remapped to keep away from LocoNet
  Port B (Pin 8 RX) on the MiniPro. Detected LocoNet errors are signaled with a beep.
  
  This project makes use of the LocoNet library developed by Alex Shepherd and the
  LiquidCrystal support library from Arduino
  =======================================================================*/
#include <LocoNet.h>
#define RECALL_STACK_LEN  6
#define AJS_RECALL_STACK  { 38, 44, 53, 99, 191, 192 }
//  PIN 8 - RX, 6 - TX TO LOCONET INTERFACE
typedef struct {
  word  RecallStack[ RECALL_STACK_LEN ] ;
  word  LastAddress ;
  byte  LastSlot ;
} 
TH_EE_CONFIG ;
TH_EE_CONFIG    eeConfig __attribute__((section(".eeprom")))
  = { AJS_RECALL_STACK, 0, 0xFF };
LocoNetThrottleClass  Throttle ;
word LocoAddr ;
byte RecallIndex ;
LnBuf LnRxBuffer ;
lnMsg *RxPacket ;
uint32_t LastThrottleTimerTick;
int onFctValue[9] = {16,1,2,4,8,1,2,4,8};    //these are the values returned when ON (0 for OFF)
int offFctValue = 0;
// ============================ Display =======================
#include <LiquidCrystal.h>
LiquidCrystal lcd(15, 255, 2, 3, 4, 17, 16); //LCD shield remapped pins
//lcd(rs,rw,en,d4,d5,d6,d7) where rw not needed
int x;                          // for working screen location
int buttonValue ;

// the 8 arrays that form each segment of the custom numbers
byte LT[8] = {B00111, B01111, B11111, B11111, B11111, B11111, B11111, B11111}; //cust char @ loc 128
byte UB[8] = {B11111, B11111, B11111, B00000, B00000, B00000, B00000, B00000};
byte RT[8] = {B11100, B11110, B11111, B11111, B11111, B11111, B11111, B11111};
byte LL[8] = {B11111, B11111, B11111, B11111, B11111, B11111, B01111, B00111};
byte LB[8] = {B00000, B00000, B00000, B00000, B00000, B11111, B11111, B11111};
byte LR[8] = {B11111, B11111, B11111, B11111, B11111, B11111, B11110, B11100};
byte UMB[8] = {B11111, B11111, B11111, B00000, B00000, B00000, B11111, B11111};
byte LMB[8] = {B11111, B00000, B00000, B00000, B00000, B11111, B11111, B11111};
// ============== Operational parameters =========================
int counterhours = 1;         //maintained in 24 hours format (1-24)
int counterminutes = 0;       //
int workhours;
char workampm;
uint16_t curAddr;             // currently selected LocoNet address
uint8_t curFunc;
uint8_t curFState;
uint8_t curAction; 
long lastTime = 0;
long lndelaytime = 1000;        // pauses to allow LocoNet time to work
long tickTime = 500;            // Action clock ticks TickTime = 500 (fastest) to 10000 (slowest)
boolean LNBusy;                 // prevent clock ticks if doing sequencial LocoNet work
uint8_t clkPtr;
#define tonePin 7               // pin 9 used for tone alert
#define fOn 1
#define fOff 0
struct Schedule {
  uint16_t  clkHours;           // time of day hours (0-23)
  uint16_t  clkMinutes;         // time of day minutes
  uint16_t  Addr;               // MFAD decoder address
  uint8_t   Func;               // MFAD function to be modified
  uint8_t   fStatus;            // fOn or fOff (1 or 0)   
};
Schedule Event [] = {
{1,30,1000,3,fOn},  //store light ON (new lnaddr adds addl 5 clock ticks)
{1,40,1000,8,fOn},  //shades UP
{1,50,1000,1,fOn},  //van flasher ON
{2,0,1000,7,fOn},   //customers talking ON
{2,25,1000,1,fOff}, //van flasher OFF
{2,35,1000,4,fOn},  //cafe sign ON
{2,55,1000,7,fOff}, //customers talking OFF
{3,5,1000,2,fOn},   //upstairs lights ON
{3,20,1000,8,fOff}, //shades DOWN
{3,30,1000,3,fOff}, //store lights OFF
{3,40,1000,0,fOn},  //TV flicker ON
{3,50,1000,4,fOff}, //cafe sign OFF
{4,0,1000,2,fOff},  //upstairs lighs OFF
{4,25,1000,0,fOff}, //TV flicker OFF
{5,0,1001,1,fOn},   //truck flasher ON  (new lnaddr adds addl 5 clock ticks)
{5,15,1001,3,fOn},  //office lights ON
{5,30,1001,2,fOn},  //shop lights ON
{5,40,1001,8,fOn},  //garage door UP
{5,55,1001,7,fOn},  //shop sounds ON
{6,10,1001,1,fOff}, //truck flasher OFF
{6,35,1001,7,fOff}, //shop sounds OFF
{6,40,1001,0,fOn},  //welding ON
{7,10,1001,0,fOff}, //welding OFF
{7,15,1001,4,fOn},  //outside signs ON
{7,30,1001,8,fOff}, //garage door DOWN
{7,50,1001,2,fOff}, //shop lights OFF
{8,0,1001,3,fOff}, //office lights OFF
{8,15,1001,4,fOff},  //outside signs OFF
{999,9,9,9,9},           //end-of-table marker, force day over
};

// ================= Setup ===================================
void setup()  {
  LocoNet.init();                 // First initialize the LocoNet interface
  Throttle.init(0, 0, 9999);
  curAddr = 0;
  //Serial.begin(57600);
  //Serial.println("Start");
  // Loads LCD custom character memory with above designs
  lcd.createChar(8,LT);
  lcd.createChar(1,UB);
  lcd.createChar(2,RT);
  lcd.createChar(3,LL);
  lcd.createChar(4,LB);
  lcd.createChar(5,LR);
  lcd.createChar(6,UMB);
  lcd.createChar(7,LMB);
  lcd.begin(0, 2);              // sets the LCD's row and column dimensions
  showSpeed();
  initScreen();
  initDay();                    // initialize screen and working parameters for new day
  Throttle.releaseAddress();
  while (analogRead(A0) < 492 || analogRead(A0) > 512) {};   //hold till Start Button
}
// ================= Time Keeper for LocoNet ===================
boolean isTime(unsigned long *timeMark, unsigned long timeInterval) {
    unsigned long timeNow = millis();
    if ( timeNow - *timeMark >= timeInterval) {
        *timeMark = timeNow;
        return true;
    }    
    return false;
}
// ===================== Main Loop ==============================
void loop() { 
    
  RxPacket = LocoNet.receive(); // Check for any received LocoNet packets
  if( RxPacket ) {
    if( !LocoNet.processSwitchSensorMessage(RxPacket) ) 
      Throttle.processMessage(RxPacket) ; 
    }  
    if ((millis() > lastTime + tickTime)) {  // OK to do another action
      lastTime = millis();

      counterminutes++;
      if (counterminutes >= 60) {
        counterminutes = 0;
        counterhours++;
      }
      if (counterhours > 24) {
         counterhours = 1;
         //Serial.println("NEXT DAY");
         initDay();                         // initialize everything for a new day 
      }

      if ((Event[clkPtr].clkMinutes == counterminutes) && (Event[clkPtr].clkHours == counterhours) && !LNBusy){  //found scheduled table item
        if (Event[clkPtr].Addr == curAddr){
          curAction = 6;          // already had address so must be a function activity
        } else { 
          curAddr = Event[clkPtr].Addr;
          curAction = 1;
        }
      }
      if ((Event[clkPtr].clkHours == 999) && !LNBusy){ 
        initDay();
      }
      
      
      //Serial.print(counterhours);
      //Serial.print(":");
      //Serial.println(counterminutes);
      doDisplay();
      doAction();
    } 
    if (!LNBusy && curAction == 0) getButtons();    //look at buttons if NOT doing LN business
    if(isTime(&LastThrottleTimerTick, 100)) {
    Throttle.process100msActions() ; 
  }
}  

void doAction() {
  switch (curAction) {
    case 1:
      Throttle.releaseAddress();
      //Serial.println("Release");
      curAction++;
      break;    
    case 2:       //release address
      //Serial.print("Freeing ");
      //Serial.println(curAddr);
      Throttle.freeAddress(curAddr);
      curAction++;
      break;
    case 3:       //locoNet activity delay
      //Serial.println("2-lag");
      delay(lndelaytime);
      curAction++;
      break;
    case 4:       // acquire new address
      //Serial.print("3-Acquire ");
      //Serial.println(curAddr);
      Throttle.setAddress(curAddr);
      curAction++;
      break;
    case 5:       //locoNet activity delay
      //Serial.println("4-lag");
      delay(lndelaytime);
      curAction++;
      break;
    case 6:       // function activity
      //Serial.print("5-Fct ");
      curFunc = Event[clkPtr].Func;
      //Serial.print(curFunc);
      curFState = Event[clkPtr].fStatus;            // fOn or fOff (1 or 0) 
      if (curFState) {
        Throttle.setFunction(curFunc, onFctValue[curFunc]);
        //Serial.println(" ON");
      } else {
        Throttle.setFunction(curFunc, offFctValue);
        //Serial.println(" OFF");
      }
      curAction++;
      break;
    case 7:                     //move on to next table entry
      //Serial.println("next table entry");
      LNBusy = false;           //let clock start up again
      curAction = 0;
      clkPtr++;
      break;
    }
}  

void getButtons() {
    buttonValue = analogRead(A0);
    if (buttonValue >=492 && buttonValue <=512) {  //around base of 502 (LEFT Button)
      //Serial.println("RESTART DAY");
      initDay();                         // initialize everything for a new day
    }
    if (buttonValue >=132 && buttonValue <=152) {  //around base of 142 (UP Button)
      tickTime = tickTime - 1000;         //lower by 1,000s (make faster clock) 
      if (tickTime < 500) tickTime = 500;
      //Serial.print("Tick Time =");
      //Serial.println(tickTime);
      showSpeed();
    }
    if (buttonValue >=318 && buttonValue <=338) {  //around base of 328 (DOWN Button)
      tickTime = tickTime + 1000;         //raise by 1,000s (make slower clock) 
      if (tickTime > 10000) tickTime = 10000;
      //Serial.print("Tick Time =");
      //Serial.println(tickTime);
      showSpeed();
    }
    if (buttonValue <=10) {             //around base of 0 (RIGHT Button) Stop/Start
      while (analogRead(A0) <492 || analogRead(A0) > 512) {};   //hold till NOT Start Button
    }
    while (analogRead(A0) < 1000) {}     //debounce by waiting for "no button" value 
}
/* Note analog values for A0 read from LCD shield are:
 *  no button = 1023
 *  left button = 502
 *  right button = 0
 *  up button = 142
 *  down button = 328
 *  select button = 728
 *  a "hysterisis" (+/-) value of 10 seems to work fine.
 */
void initScreen() {
  lcd.clear();
  lcd.setCursor(7,0);    // place colon
  lcd.write(165);
  lcd.setCursor(7,1);
  lcd.write(165);
  lcd.setCursor(15,0);
  lcd.print("A");
  lcd.setCursor(15,1);
  lcd.print("M");
}
void initDay(){
  clkPtr = 0;
  LNBusy = false;
  counterhours = 1;
  counterminutes = 0;
  doDisplay();
}

// ============== LOCONET CALL BACK AND SUPPORT FUNCTIONS =========

void notifyThrottleError( uint8_t UserData, TH_ERROR Error ) {
  //Serial.print("Error ");
  //Serial.print(Error, DEC);
  //Serial.print(' ');
  //Serial.println(Throttle.getErrorStr(Error));
  tone (tonePin,1000,500);    //sound off alert
}
// ====================== Display Stuff ==============================

void displayNumbers(byte myNumber) {
  switch (myNumber) {
    case 0:
      lcd.setCursor(x, 0);
      lcd.write(8);
      lcd.write(1);
      lcd.write(2);
      lcd.setCursor(x, 1); 
      lcd.write(3);  
      lcd.write(4);  
      lcd.write(5);
      break;
    case 1:
      lcd.setCursor(x,0);
      lcd.write(32);
      lcd.write(255);
      lcd.write(32);
      lcd.setCursor(x,1);
      lcd.write(32);
      lcd.write(255);
      lcd.write(32);
      break;
    case 2:
      lcd.setCursor(x,0);
      lcd.write(6);
      lcd.write(6);
      lcd.write(2);
      lcd.setCursor(x, 1);
      lcd.write(3);
      lcd.write(7);
      lcd.write(7);
      break;
    case 3:
      lcd.setCursor(x,0);
      lcd.write(6);
      lcd.write(6);
      lcd.write(2);
      lcd.setCursor(x, 1);
      lcd.write(7);
      lcd.write(7);
      lcd.write(5);
      break;
    case 4:
      lcd.setCursor(x,0);
      lcd.write(3);
      lcd.write(4);
      lcd.write(2);
      lcd.setCursor(x, 1);
      lcd.write(32);
      lcd.write(32);
      lcd.write(255);
      break;
    case 5:
      lcd.setCursor(x,0);
      lcd.write(255);
      lcd.write(6);
      lcd.write(6);
      lcd.setCursor(x, 1);
      lcd.write(7);
      lcd.write(7);
      lcd.write(5);
      break;
    case 6:
      lcd.setCursor(x,0);
      lcd.write(8);
      lcd.write(6);
      lcd.write(6);
      lcd.setCursor(x, 1);
      lcd.write(3);
      lcd.write(7);
      lcd.write(5);
      break;
    case 7:
      lcd.setCursor(x,0);
      lcd.write(1);
      lcd.write(1);
      lcd.write(2);
      lcd.setCursor(x, 1);
      lcd.write(32);
      lcd.write(8);
      lcd.write(32);
      break;
    case 8:
      lcd.setCursor(x,0);
      lcd.write(8);
      lcd.write(6);
      lcd.write(2);
      lcd.setCursor(x, 1);
      lcd.write(3);
      lcd.write(7);
      lcd.write(5);
      break;
    case 9:
      lcd.setCursor(x,0);
      lcd.write(8);
      lcd.write(6);
      lcd.write(2);
      lcd.setCursor(x, 1);
      lcd.write(7);
      lcd.write(7);
      lcd.write(5);
      break;
    case 10:                // This is a blank character
      lcd.setCursor(x,0);
      lcd.write(32);
      lcd.write(32);
      lcd.write(32);
      lcd.setCursor(x, 1);
      lcd.write(32);
      lcd.write(32);
      lcd.write(32);
      break;
  }
  x = x+3;
}

void doDisplay(){
    x = 1;                                    //left position, 1st digit
    workhours = counterhours;                 //copy of 24 hour verson
    if (workhours > 12 && workhours < 24) {
      workhours = workhours - 12;             //convert to 12 hour version
      workampm = 'P';
        }else{
      workampm = 'A';
    }
    if (workhours == 24) workhours = 12;
    if (counterhours == 12 ) workampm = 'P';
    if (counterhours == 24 ) workampm = 'A'; 
    displayNumbers(workhours / 10);
    displayNumbers(workhours % 10);
    x++;                                      //pass over colon
    displayNumbers(counterminutes / 10);
    displayNumbers(counterminutes % 10);
    lcd.setCursor(15,0);
    lcd.print(workampm);
}
void showSpeed(){
      lcd.clear();
      lcd.print("CLOCK: ");
      lcd.print(60000/tickTime);
      lcd.print('X');
      delay (3000);                   //hold display for 3 seconds
      initScreen();
      doDisplay();
}
